import csv
import json
import shutil
import zipfile
from pprint import pformat
from cStringIO import StringIO

from crowdflower import logger
from crowdflower.exception import CrowdFlowerError
from crowdflower.cache import cacheable, keyfunc
from crowdflower.serialization import rails_params


class Job(object):
    '''
    Read / Write attributes
        auto_order
        auto_order_threshold
        auto_order_timeout
        cml
        cml_fields
        confidence_fields
        css
        custom_key
        excluded_countries
        gold_per_assignment
        included_countries
        instructions
        js
        judgments_per_unit
        language
        max_judgments_per_unit
        max_judgments_per_contributor
        min_unit_confidence
        options
        pages_per_assignment
        problem
        send_judgments_webhook
        state
        title
        units_per_assignment
        webhook_uri

    Read-only attributes
        completed
        completed_at
        created_at
        gold
        golds_count
        id
        judgments_count
        units_count
        updated_at

    Not sure about:
        payment_cents

    '''
    READ_WRITE_FIELDS = ['auto_order', 'auto_order_threshold', 'auto_order_timeout', 'cml', 'cml_fields', 'confidence_fields', 'css', 'custom_key', 'excluded_countries', 'gold_per_assignment', 'included_countries', 'instructions', 'js', 'judgments_per_unit', 'language', 'max_judgments_per_unit', 'max_judgments_per_contributor', 'min_unit_confidence', 'options', 'pages_per_assignment', 'problem', 'send_judgments_webhook', 'state', 'title', 'units_per_assignment', 'webhook_uri']
    _cache_key_attrs = ('id',)

    def __init__(self, job_id, connection):
        self.id = job_id
        self._connection = connection
        self._cache = self._connection._cache

    def __repr__(self):
        return pformat(self.properties)

    @property
    @cacheable()
    def properties(self):
        return self._connection.request('/jobs/%s' % self.id)


    @cacheable('tags')
    def get_tags(self):
        res = self._connection.request('/jobs/%s/tags' % self.id)
        return [item['name'] for item in res]

    def set_tags(self, tags):
        params = rails_params({'tags': tags})
        self._connection.request('/jobs/%s/tags' % self.id, method='PUT', params=params)
        self._cache.remove(keyfunc(self, 'tags'))

    tags = property(get_tags, set_tags)

    def add_tags(self, tags):
        params = rails_params({'tags': tags})
        self._connection.request('/jobs/%s/tags' % self.id, method='POST', params=params)
        self._cache.remove(keyfunc(self, 'tags'))

    @property
    @cacheable()
    def units(self):
        '''
        Returns a list of units, e.g.,

            [
                {
                    '_unit_id': '495781935',
                    'id': 'may25_1029',
                    'text': 'remember when I was in hospital for four months nd it was my birthday nd everyone forgot nd no one even came to visit me'
                },
                {
                    '_unit_id': '495781936',
                    'id': 'may25_1030',
                    'text': 'I had the wifi taken away and I can't have any friends over what a great Summer!'
                }
                ...
            ]

        N.b.: this is not the same format that CrowdFlower responds with, which is more like:

            {
                '495781935': {
                    'id': 'may25_1029',
                    'text': 'remember when I was in hospital for four months nd it was my birthday nd everyone forgot nd no one even came to visit me'
                },
                '495781936': {
                    'id': 'may25_1030',
                    'text': 'I had the wifi taken away and I can't have any friends over what a great Summer!'
                }
                ...
            }

        But since we are caching and de-paginating, it's better to return a list (by being an iterator)

        Neither `page` nor `limit` query parameters are documented, but they work as expected. Page numbering starts at 1, and the limit has a maximum of 1000 (which is also the default).

        The returned units are not fully hydrated; in fact, the response only includes each unit id and the original unit payload. To see the crowd's responses, you must use /jobs/{job.id}/units/{unit.id}, or the bulk download method (see Job.download()).

        '''
        page = 0
        while True:
            page += 1
            params = dict(page=page)
            units_response = self._connection.request('/jobs/%s/units' % self.id, params=params)
            for unit_id, unit_properties in units_response.items():
                # hopefully the user has not specified '_unit_id' as a custom field
                unit_properties['_unit_id'] = unit_id
                yield unit_properties
            if len(units_response) < 1000:
                break


    def delete_unit(self, unit_id):
        response = self._connection.request('/jobs/%s/units/%s' % (self.id, unit_id), method='DELETE')
        # bust cache if the request did not raise any errors
        self._cache.remove(keyfunc(self, 'units'))
        return response

    def upload(self, units):
        headers = {'Content-Type': 'application/json'}
        data = '\n'.join(json.dumps(unit) for unit in units)
        logger.debug('Uploading data to Job[%d]: %s', self.id, data)
        res = self._connection.request('/jobs/%s/upload' % self.id, method='POST', headers=headers, data=data)

        # reset cached units
        self._cache.remove(keyfunc(self, 'units'))

        return res

    def update(self, props):
        logger.debug('Updating Job[%d]: %r', self.id, props)
        params = rails_params({'job': props})

        try:
            res = self._connection.request('/jobs/%s' % self.id, method='PUT', params=params)
        except CrowdFlowerError, exc:
            # CrowdFlower sometimes likes to redirect the PUT to a non-API page,
            # which will raise an error (406 Not Accepted), but we can just
            # ignore the error since it comes after the update is complete.
            # This is kind of a hack, since we sometimes want to follow redirects
            # (e.g., with downloads), but following redirects is more properly
            # not the default
            if exc.response.status_code != 406:
                logger.info('Ignoring 406 "Not Accepted" error: %r', exc);
            else:
                raise

        # reset cached properties
        self._cache.remove(keyfunc(self, 'properties'))

        return res

    def channels(self):
        '''
        Manual channel control is deprecated.

        The API documentation includes a PUT call at this endpoint, but I'm
        not sure if it actually does anything.
        '''
        return self._connection.request('/jobs/%s/channels' % self.id)

    def legend(self):
        '''
        From the CrowdFlower documentation:

        > The legend will show you the generated keys that will end up being
        > submitted with your form.
        '''
        return self._connection.request('/jobs/%s/legend' % self.id)

    def gold_reset(self):
        '''
        Mark all of this job's test questions (gold data) as NOT gold.

        Splitting the /jobs/:job_id/gold API call into gold_reset() and
        gold_add() is not faithful to the API, but resetting gold marks
        and adding them should not have the same API endpoint in the first place.
        '''
        params = dict(reset='true')
        res = self._connection.request('/jobs/%s/gold' % self.id, method='PUT', params=params)
        # reset cache
        self._cache.remove(keyfunc(self, 'properties'))
        self._cache.remove(keyfunc(self, 'units'))
        return res

    def gold_add(self, check, check_with=None):
        '''
        Configure the gold labels for a task.

        * check: the name of the field being checked against
            - Can call /jobs/{job_id}/legend to see options
            - And as far as I can tell, the job.properties['gold'] field is a
              hash with keys that are "check" names, and values that are "with" names.
        * check_with: the name of the field containing the gold label for check
            - Crowdflower calls this field "with", which is a Python keyword
            - defaults to check + '_gold'

        I'm not sure why convert_units would be anything but true.
        '''
        params = dict(check=check, convert_units='true')
        if check_with is not None:
            params['with'] = check_with
        res = self._connection.request('/jobs/%s/gold' % self.id, method='PUT', params=params)
        # reset cache
        self._cache.remove(keyfunc(self, 'properties'))
        self._cache.remove(keyfunc(self, 'units'))
        return res

    def launch(self, units_count, channels=('on_demand',)):
        '''
        `units_count` is the number of units to order
        `channels` should be some non-empty combination of 'cf_internal'
            (sandbox mode) and / or 'on_demand' (normal)
        '''
        channels = list(channels)
        data = rails_params(dict(channels=channels, debit=dict(units_count=units_count)))
        res = self._connection.request('/jobs/%s/orders' % self.id, method='POST', params=data)
        self._cache.remove(keyfunc(self, 'properties'))
        return res

    def ping(self):
        '''
        Determine the status of a job.

        Returns a data structure like this:
            {
                "all_units": 298,
                "golden_units": 239,
                "ordered_units": 49,
                "all_judgments": 168,
                "needed_judgments": 0,
                "tainted_judgments": 45,
                "completed_gold_estimate": 239,
                "completed_non_gold_estimate": 49,
                "completed_units_estimate": 288
            }
        '''
        return self._connection.request('/jobs/%s/ping' % self.id)

    def delete(self):
        '''
        Deletes the entire job permanently
        '''
        return self._connection.request('/jobs/%s' % self.id, method='DELETE')

    def download(self, full=True):
        '''The resulting CSV will have headers like:

            _unit_id
                Integer
                Unique ID per unit
            _created_at
                Date: m/d/yyyy hh:mm:ss
            _golden
                Enum: "true" | "false"
            _canary
                Always empty, ???
            _id
                Integer
                Unique ID per judgment
            _missed
                ???
            _started_at
                Date: m/d/yyyy hh:mm:ss
                Can use
            _tainted
                Always false, ???
            _channel
                Enum: "neodev" | "clixsense" | [etc.]
            _trust
                Always 1, ???
            _worker_id
                Integer
                Unique ID per worker
            _country
                3-letter ISO code
            _region
                String
                A number for all countries except UK, USA, Canada (others?)
            _city
                String
                City name
            _ip
                String
                IPv4 address

        And then the rest just copies over whatever fields were originally used, e.g.:

            id
            text
            sentiment
            sentiment_gold
        '''
        # pulls down the csv endpoint, unzips it, and returns a list of all the rows
        params = dict(full='true' if full else 'false')
        # use .csv, not headers=dict(Accept='text/csv'), which Crowdflower rejects
        req = self._connection.create_request('/jobs/%s.csv' % self.id, method='GET', params=params)
        res = self._connection.send_request(req)
        # because ZipFile insists on seeking, we can't simply pass over the res.raw stream
        fp = StringIO()
        fp.write(res.content)
        # ZipFile does fp.seek(0) itself
        zf = zipfile.ZipFile(fp)
        for zipinfo in zf.filelist:
            zipinfo_fp = zf.open(zipinfo)
            reader = csv.DictReader(zipinfo_fp)
            for row in reader:
                yield {key: value.decode('utf8') for key, value in row.items()}


    @property
    @cacheable()
    def judgments(self):
        return self.download()


    def download_csv(self, filepath, full=True):
        '''
        Basically the same as job.judgments but without parsing the csv.
        '''
        params = dict(full='true' if full else 'false')
        req = self._connection.create_request('/jobs/%s.csv' % self.id, method='GET', params=params)
        res = self._connection.send_request(req)
        fp = StringIO()
        fp.write(res.content)
        zf = zipfile.ZipFile(fp)
        zipinfo = zf.filelist[0]
        fsrc = zf.open(zipinfo)
        with open(filepath, 'w') as fdst:
            shutil.copyfileobj(fsrc, fdst)



import os
import re
import json
import unicodedata
from crowdflower import logger


def clean_filename(filename):
    '''
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.

    From Django, partly, via http://stackoverflow.com/q/295135/424651

    Django is BSD licensed, so, it's cool.
    '''
    filename_unicode = unicode(filename)
    filename_normalized = unicodedata.normalize('NFKD', filename_unicode)
    # reduce to ascii charset
    filename_ascii = filename_normalized.encode('ascii', 'ignore')
    # delete everything that's not a word character, whitespace, _, or -
    filename_cleaned = re.sub(r'[^\w\s_-]', '', filename_ascii)
    # turn all sequences of whitespace to a single space
    filename_collapsed = re.sub(r'\s+', ' ', filename_cleaned).strip()
    # remove leading hyphens (less painful in Linux)
    filename_nice = filename_collapsed.lstrip('-')
    return filename_nice


def flatten(obj):
    if hasattr(obj, '__iter__'):
        # dicts are iterable, but we don't want to flatten them into a list
        if not isinstance(obj, dict):
            return list(obj)
    return obj


def keyfunc(instance, func_attr):
    cache_key_values = [getattr(instance, cache_key_attr) for cache_key_attr in instance._cache_key_attrs]
    return '%s[%s].%s' % (instance.__class__.__name__, ':'.join(map(str, cache_key_values)), func_attr)


def cacheable(name=None):
    '''An object method decorator. Use like:

    class User(object):
        _cache_key_attrs = ('user_id',)

        def __init__(self, user_id):
            self.user_id = user_id
            self._cache = FilesystemCache()

        @cacheable()
        def actions(self):
            yield 'logged in'
            yield 'changed password'
            yield 'searched for somethign'
            # etc.

        @cacheable('tags')
        def get_tags(self):
            return ['a', 'b']

        tags = property(get_tags)

    The class it is used in must have a `_cache_key_attrs` attribute and a `_cache` Cache instance.
    '''
    def decorator(func):
        func_attr = name or func.__name__
        def wrapper(self, *args, **kwargs):
            key = keyfunc(self, func_attr)
            value = self._cache.get(key)
            if value is None:
                logger.info('cache miss; fetching "%s" and writing to cache', key)
                # self refers to the instance, which SHOULD have a ._cache attribute
                value = func(self, *args, **kwargs)
                value = flatten(value)
                self._cache.put(key, value)
            else:
                logger.info('cache hit; reading "%s" from cache', key)
            return value
        return wrapper
    return decorator


class AbstractCache(object):
    def get(self, key):
        raise NotImplementedError

    def put(self, key, value):
        raise NotImplementedError

    def remove(self, key):
        raise NotImplementedError

    def removeAll(self):
        raise NotImplementedError


class NoCache(AbstractCache):
    def get(self, key):
        return None

    def put(self, key, value):
        pass

    def remove(self, key):
        pass

    def removeAll(self):
        pass


class FilesystemCache(AbstractCache):
    def __init__(self, dirpath='/tmp/crowdflower'):
        self.dirpath = dirpath
        if not (os.path.exists(dirpath) and os.path.isdir(dirpath)):
            # this should rightly fail if dirpath is a file, or cannot be accessed
            os.makedirs(dirpath)

    def _filename(self, key):
        return os.path.join(self.dirpath, clean_filename(key)) + '.json'

    def get(self, key):
        filepath = self._filename(key)
        if os.path.exists(filepath):
            with open(filepath) as fp:
                return json.load(fp)

    def put(self, key, value):
        filepath = self._filename(key)
        with open(filepath, 'w') as fp:
            json.dump(value, fp)

    def remove(self, key):
        filepath = self._filename(key)
        if os.path.exists(filepath):
            os.remove(filepath)

    def removeAll(self):
        for filename in os.listdir(self.dirpath):
            filepath = os.path.join(self.dirpath, filename)
            os.remove(filepath)



import os
import json
from crowdflower.exception import CrowdFlowerError, CrowdFlowerJSONError
from requests import Request, Session
# I regret that python-requests can't handle merging lists of params
from requests.utils import to_key_val_list

from crowdflower import logger
from crowdflower.job import Job
from crowdflower.cache import FilesystemCache, NoCache, cacheable, keyfunc


class Connection(object):
    DEFAULT_API_KEY = os.getenv('CROWDFLOWER_API_KEY')
    DEFAULT_API_URL = 'https://api.crowdflower.com/v1'
    _cache_key_attrs = ('api_key',)

    def __init__(self, cache=None, api_key=DEFAULT_API_KEY, api_url=DEFAULT_API_URL):
        self.api_key = api_key
        self.api_url = api_url

        if cache == 'filesystem':
            self._cache = FilesystemCache()
        # elif ... others?
        else:
            self._cache = NoCache()

        self._session = Session()
        # self._session.params['key'] = self.api_key
        # self._session.verify = False


    def __repr__(self):
        return '<{:} using {:} with key {:}...>'.format(self.__class__.__name__, self.api_url, self.api_key[:6])


    def create_request(self, path, method='GET', **kw):
        url = self.api_url + path
        return Request(method=method, url=url, **kw)

    def send_request(self, req):
        '''
        returns requests.Response object

        raise
        '''
        # requests gotcha: even if send through the session, request.prepare()
        # does not get merged with the session's attributes, so we have to call
        # session.prepare_request(...)

        # merge params with the api key
        req.params = to_key_val_list(req.params) + [('key', self.api_key)]
        prepared_req = self._session.prepare_request(req)
        logger.debug('Request params: {}'.format(req.params))

        res = self._session.send(prepared_req)
        if res.status_code != 200:
            # CrowdFlower responds with a '202 Accepted' when we request a bulk
            # download which has not yet been generated, which means we simply
            # have to wait and try again
            raise CrowdFlowerError(req, res)
        return res

    def request(self, path, method='GET', params=None, headers=None, data=None):
        # simple request helper
        if headers is None:
            headers = dict()
        headers.update(Accept='application/json')
        req = self.create_request(path, method=method, params=params, headers=headers, data=data)
        res = self.send_request(req)
        try:
            # what Requests might actually raise is a simplejson.scanner.JSONDecodeError,
            # but I'm pretty sure that's the only error .json() might raise, so we don't
            # to type-match it.
            return res.json()
        except Exception, err:
            raise CrowdFlowerJSONError(req, res, err)

    def job(self, job_id):
        # doesn't actually call anything
        return Job(job_id, self)

    @property
    @cacheable()
    def job_ids(self):
        '''
        The API documentation does not specify this, but there is a hard-coded
        limit=10 parameter on the /jobs endpoint, and no total count, so we must
        page through until we get a response with fewer than 10 items.

        Apparently, other parameters, like query='pt' and fields[]='tags' don't work.
        '''
        page = 0
        while True:
            page += 1
            params = dict(page=page)
            jobs_response = self.request('/jobs', params=params)
            for job_properties in jobs_response:
                # somehow add the Job's properties to the cache, since we have all the data anyway?
                yield job_properties['id']
            if len(jobs_response) < 10:
                break

    def jobs(self):
        '''
        This is separated from job_ids to aid in caching. They have no cause
        to be separate except to avoid marshaling issues when encoding/decoding.
        '''
        for job_id in self.job_ids:
            yield Job(job_id, self)

    def upload(self, units):
        '''
        TODO: allow setting Job parameters at the same time
        '''
        headers = {'Content-Type': 'application/json'}
        # N.b.: CF expects newline-separated JSON, not actual JSON
        # e.g., this would fail with a status 500: kwargs['data'] = json.dumps(data)
        data = '\n'.join(json.dumps(unit) for unit in units)

        job_response = self.request('/jobs/upload', method='POST', headers=headers, data=data)
        job = Job(job_response['id'], self)
        job._properties = job_response
        # bust cache of job_ids
        self._cache.remove(keyfunc(self, 'job_ids'))
        return job

    def account(self):
        '''
        This is very short and simple, but it's not documented in the API.
        '''
        return self.request('/account')



def rails(value, prefix=''):
    if isinstance(value, list):
        for subvalue in value:
            for pair in rails(subvalue, prefix=prefix + '[]'):
                yield pair
    elif isinstance(value, dict):
        for subkey, subvalue in value.items():
            for pair in rails(subvalue, prefix=prefix + '[' + subkey + ']'):
                yield pair
    else:
        yield prefix, value


def rails_params(params):
    for root, value in params.items():
        for pair in rails(value, prefix=root):
            yield pair



import os
# here is the directory containing this __init__.py file
here = os.path.dirname(__file__) or os.curdir
# root is the directory containing `here`, i.e., the directory containing setup.py
root = os.path.dirname(os.path.abspath(here))


import pkg_resources
__version__ = pkg_resources.get_distribution('crowdflower').version


import logging
logger = logging.getLogger('crowdflower')


# import helper
from connection import Connection



from setuptools import setup, find_packages

setup(
    name='crowdflower',
    version='0.1.1',
    author='Christopher Brown',
    author_email='io@henrian.com',
    url='https://github.com/peoplepattern/crowdflower',
    keywords='crowdflower crowdsourcing api client',
    description='CrowdFlower API - Python client',
    long_description=open('README.rst').read(),
    license=open('LICENSE').read(),
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        # https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
    ],
    install_requires=[
        'requests>=2.0.0',
    ],
)



class Unit(object):
    '''
    Read / Write attributes
        job_id
        missed_count
        difficulty
        state
        data
        agreement

    Read-only attributes
        updated_at
        created_at
        judgments_count
        id
    '''



class CrowdFlowerError(Exception):
    """
    General CrowdFlower error
    """
    def __init__(self, request, response):
        self.request = request
        self.response = response

    def __str__(self):
        return '%s: %d %s at %s' % (
            self.__class__.__name__,
            self.response.status_code,
            self.response.reason,
            self.request.url)

    __repr__ = __str__


class CrowdFlowerJSONError(CrowdFlowerError):
    """
    CrowdFlower JSON parsing error.

    This is raised when JSON is expected from CrowdFlower API,
    but fails to parse for whatever reason.
    """
    def __init__(self, request, response, json_error):
        self.request = request
        self.response = response
        self.json_error = json_error
